@@ -69,6 +69,8 @@ static void handle_iq_command_leave_groupchats(jconn conn, char *from,
 
 typedef void (*adhoc_command_callback)(jconn, char*, const char*, xmlnode);
 
+inline double seconds_since_last_use(void);
+
 struct adhoc_command {
   char *name;
   char *description;
@@ -1147,6 +1149,7 @@ static void handle_iq_command_set_status(jconn conn, char *from, const char *id,
     sessionid = generate_session_id("set-status");
     xmlnode_put_attrib(command, "sessionid", sessionid);
     g_free(sessionid);
+    sessionid = NULL;
     xmlnode_put_attrib(command, "status", "executing");
 
     x = xmlnode_insert_tag(command, "x");
@@ -1187,12 +1190,12 @@ static void handle_iq_command_set_status(jconn conn, char *from, const char *id,
     // I do not think this is useful, user should not have to care of the
     // priority like gossip and gajim do (misc)
     y = xmlnode_insert_tag(x, "field");
-    xmlnode_put_attrib(y, "type", "text-multi");
+    xmlnode_put_attrib(y, "type", "text-single");
     xmlnode_put_attrib(y, "var", "status-message");
     xmlnode_put_attrib(y, "label", "Message");
-  }
-  else // (if sessionid)
-  {
+  } else if (action && !strcmp(action, "cancel")) {
+    xmlnode_put_attrib(command, "status", "canceled");
+  } else  { // (if sessionid and not canceled)
     y = xmlnode_get_tag(x, "x?xmlns=jabber:x:data");
     if (y) {
       char *value, *message;
@@ -1200,19 +1203,23 @@ static void handle_iq_command_set_status(jconn conn, char *from, const char *id,
                                    "value");
       message = xmlnode_get_tag_data(xmlnode_get_tag(y,
                                    "field?var=status-message"), "value");
-      for (s = adhoc_status_list; !s->name || strcmp(s->name, value); s++);
-      if (s->name) {
-        char *status = g_strdup_printf("%s %s", s->status,
-                                       message ? message : "");
-        cmd_setstatus(NULL, status);
-        g_free(status);
-        xmlnode_put_attrib(command, "status", "completed");
-        xmlnode_put_attrib(iq, "type", "result");
-        xmlnode_insert_dataform_result_message(command,
-                                               "Status has been changed");
+      if (value) {
+        for (s = adhoc_status_list; s->name && strcmp(s->name, value); s++);
+        if (s->name) {
+          char *status = g_strdup_printf("%s %s", s->status,
+                                         message ? message : "");
+          cmd_setstatus(NULL, status);
+          g_free(status);
+          xmlnode_put_attrib(command, "status", "completed");
+          xmlnode_put_attrib(iq, "type", "result");
+          xmlnode_insert_dataform_result_message(command,
+                                                 "Status has been changed");
+        }
       }
     }
   }
+  if (sessionid)
+    xmlnode_put_attrib(command, "sessionid", sessionid);
   xmlnode_put_attrib(iq, "to", xmlnode_get_attrib(xmldata, "from"));
   xmlnode_put_attrib(iq, "id", id);
   jab_send(jc, iq);
@@ -1262,6 +1269,7 @@ static void handle_iq_command_leave_groupchats(jconn conn, char *from,
     sessionid = generate_session_id("leave-groupchats");
     xmlnode_put_attrib(command, "sessionid", sessionid);
     g_free(sessionid);
+    sessionid = NULL;
     xmlnode_put_attrib(command, "status", "executing");
 
     x = xmlnode_insert_tag(command, "x");
@@ -1289,9 +1297,9 @@ static void handle_iq_command_leave_groupchats(jconn conn, char *from,
     xmlnode_insert_tag(field, "required");
 
     foreach_buddy(ROSTER_TYPE_ROOM, &_callback_foreach_buddy_groupchat, &field);
-  }
-  else // (if sessionid)
-  {
+  } else if (action && !strcmp(action, "cancel")) {
+    xmlnode_put_attrib(command, "status", "canceled");
+  } else  { // (if sessionid and not canceled)
     xmlnode form = xmlnode_get_tag(x, "x?xmlns=jabber:x:data");
     if (form) {
       xmlnode x, gc;
@@ -1305,7 +1313,7 @@ static void handle_iq_command_leave_groupchats(jconn conn, char *from,
         if (to_leave) {
           GList* b = buddy_search_jid(to_leave);
           if (b)
-            cmd_room_leave(b->data, "Asked by remote command");
+            cmd_room_leave(b->data, "Requested by remote command");
         }
       }
       xmlnode_put_attrib(iq, "type", "result");
@@ -1313,6 +1321,8 @@ static void handle_iq_command_leave_groupchats(jconn conn, char *from,
                                              "Groupchats have been left");
     }
   }
+  if (sessionid)
+    xmlnode_put_attrib(command, "sessionid", sessionid);
   xmlnode_put_attrib(iq, "to", xmlnode_get_attrib(xmldata, "from"));
   xmlnode_put_attrib(iq, "id", id);
   jab_send(jc, iq);
@@ -1462,7 +1472,7 @@ static void handle_iq_disco_info(jconn conn, char *from, const char *id,
   xmlnode_free(x);
 }
 
-inline double seconds_since_last_use(void)
+double seconds_since_last_use(void)
 {
   return difftime(time(NULL), iqlast);
 }
@@ -1474,8 +1484,10 @@ static void handle_iq_last(jconn conn, char *from, const char *id,
   xmlnode myquery;
   char *seconds;
 
-  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ last time request from <%s>",
-               from);
+  if (!settings_opt_get_int("iq_hide_requests")) {
+    scr_LogPrint(LPRINT_LOGNORM, "Received an IQ last time request from <%s>",
+                 from);
+  }
 
   x = jutil_iqnew(JPACKET__RESULT, NS_LAST);
   xmlnode_put_attrib(x, "id", id);
@@ -1505,8 +1517,10 @@ static void handle_iq_version(jconn conn, char *from, const char *id,
   char *os = NULL;
   char *ver = mcabber_version();
 
-  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ version request from <%s>",
-               from);
+  if (!settings_opt_get_int("iq_hide_requests")) {
+    scr_LogPrint(LPRINT_LOGNORM, "Received an IQ version request from <%s>",
+                 from);
+  }
 
   if (!settings_opt_get_int("iq_version_hide_os")) {
     struct utsname osinfo;
@@ -1544,7 +1558,9 @@ static void handle_iq_time(jconn conn, char *from, const char *id,
 
   time(&now_t);
 
-  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ time request from <%s>", from);
+  if (!settings_opt_get_int("iq_hide_requests")) {
+    scr_LogPrint(LPRINT_LOGNORM, "Received an IQ time request from <%s>", from);
+  }
 
   buf = g_new0(char, 512);
 
@@ -1591,7 +1607,9 @@ static void handle_iq_time202(jconn conn, char *from, const char *id,
 
   time(&now_t);
 
-  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ time request from <%s>", from);
+  if (!settings_opt_get_int("iq_hide_requests")) {
+    scr_LogPrint(LPRINT_LOGNORM, "Received an IQ time request from <%s>", from);
+  }
 
   buf = g_new0(char, 512);
 
@@ -1709,6 +1727,7 @@ static void handle_iq_set(jconn conn, char *from, xmlnode xmldata)
     ns = xmlnode_get_attrib(x, "xmlns");
     if (ns && !strcmp(ns, NS_COMMANDS)) {
       handle_iq_commands(conn, from, id, xmldata);
+      return;
     } else {
       iq_not_implemented = TRUE;
     }
