@@ -34,6 +34,7 @@
 #include "settings.h"
 #include "nohtml.h"
 
+#define OTR_PROTOCOL_NAME "jabber"
 
 static OtrlUserState userstate = NULL;
 static char * account = NULL;
@@ -118,6 +119,9 @@ void otr_init(const char *fjid)
 {
   char *root;
 
+  if (userstate) //already initialised
+    return;
+
   otr_is_enabled = !!settings_opt_get_int("otr");
 
   if (!otr_is_enabled)
@@ -125,7 +129,7 @@ void otr_init(const char *fjid)
 
   OTRL_INIT;
 
-  userstate = otrl_userstate_create ();
+  userstate = otrl_userstate_create();
 
   root = otr_get_dir();
   account = jidtodisp(fjid);
@@ -135,7 +139,7 @@ void otr_init(const char *fjid)
 
   if (otrl_privkey_read(userstate, keyfile)){
     scr_LogPrint(LPRINT_LOGNORM, "Could not read OTR key from %s", keyfile);
-    cb_create_privkey(NULL, account, "jabber");
+    cb_create_privkey(NULL, account, OTR_PROTOCOL_NAME);
   }
   if (otrl_privkey_read_fingerprints(userstate, fprfile, NULL, NULL)){
     scr_LogPrint(LPRINT_LOGNORM, "Could not read OTR fingerprints from %s",
@@ -197,8 +201,14 @@ static char * otr_get_dir(void)
 static ConnContext * otr_get_context(const char *buddy)
 {
   int null = 0;
-  return otrl_context_find(userstate, buddy, account, "jabber", 1, &null,
-                           NULL, NULL);
+  ConnContext * ctx;
+  char * lowcasebuddy = g_strdup(buddy);
+
+  mc_strtolower(lowcasebuddy);
+  ctx = otrl_context_find(userstate, lowcasebuddy, account, OTR_PROTOCOL_NAME,
+                          1, &null, NULL, NULL);
+  g_free(lowcasebuddy);
+  return ctx;
 }
 
 static void otr_message_disconnect(ConnContext *ctx)
@@ -356,18 +366,20 @@ int otr_receive(char **otr_data, const char * buddy, int * free_msg)
   OtrlTLV *tlv = NULL;
   ConnContext * ctx;
 
-  *free_msg = 0;
-  ignore_message = otrl_message_receiving(userstate, &ops, NULL, account,
-    "jabber", buddy, *otr_data, &newmessage, &tlvs, NULL, NULL);
-
   ctx = otr_get_context(buddy);
+  *free_msg = 0;
+  ignore_message = otrl_message_receiving(userstate, &ops, NULL,
+                                          ctx->accountname, ctx->protocol,
+                                          ctx->username, *otr_data,
+                                          &newmessage, &tlvs,NULL, NULL);
+
 
   tlv = otrl_tlv_find(tlvs, OTRL_TLV_DISCONNECTED);
   if (tlv) {
     /* Notify the user that the other side disconnected. */
     if (ctx) {
       cb_gone_insecure(NULL, ctx);
-      otr_disconnect(buddy);
+      otr_disconnect(ctx->username);
     }
   }
 
@@ -396,12 +408,17 @@ int otr_send(char **msg, const char *buddy)
   char *htmlmsg;
   ConnContext * ctx = otr_get_context(buddy);
 
-  htmlmsg = html_escape(*msg);
-
-  err = otrl_message_sending(userstate, &ops, NULL, account, "jabber", buddy,
-                             htmlmsg, NULL, &newmessage, NULL, NULL);
-
-  g_free(htmlmsg);
+  if (ctx->msgstate == OTRL_MSGSTATE_PLAINTEXT)
+    err = otrl_message_sending(userstate, &ops, NULL, ctx->accountname,
+                               ctx->protocol, ctx->username, *msg, NULL,
+                               &newmessage, NULL, NULL);
+  else {
+    htmlmsg = html_escape(*msg);
+    err = otrl_message_sending(userstate, &ops, NULL, ctx->accountname,
+                               ctx->protocol, ctx->username, htmlmsg, NULL,
+                               &newmessage, NULL, NULL);
+    g_free(htmlmsg);
+  }
 
   if (err)
     *msg = NULL; /*something went wrong, don't send the plain-message! */
@@ -421,7 +438,7 @@ void otr_print_info(const char * buddy)
 {
   const char *state, *auth, *policy;
   ConnContext * ctx = otr_get_context(buddy);
-  OtrlPolicy p = cb_policy (ctx->app_data, ctx);
+  OtrlPolicy p = cb_policy(ctx->app_data, ctx);
 
   if (!userstate || !ctx)
     return;
@@ -521,7 +538,7 @@ void otr_smp_respond(const char * buddy, const char * secret)
   if (ctx) {
     if (!ctx->smstate->secret) {
       scr_LogPrint(LPRINT_LOGNORM,
-                   "Don't call smpr before you haven't received an SMP "
+                   "Don't call smpr until you have received an SMP "
                    "Initiation!");
       return;
     }
