@@ -3,9 +3,12 @@
 
 #include <config.h>
 
+#if defined HAVE_UNICODE && defined HAVE_WCHAR_H && defined HAVE_WCTYPE_H
+# define UNICODE
+#endif
+
 #ifdef HAVE_WCHAR_H
 # include <wchar.h>
-# define UNICODE
 # define get_char_width(c) (utf8_mode ? wcwidth(get_char(c)) : 1)
 #else
 # define wcwidth(c) 1
@@ -14,15 +17,24 @@
 
 #ifdef HAVE_WCTYPE_H
 # include <wctype.h>
+
+/* The following bit is a hack for Solaris 8&9 systems that don't have
+ * iswblank().
+ * For now i made sure it comes after wctype.h so it doesn't create
+ * problems (wctype.h has calls to iswblank() before wctype() is declared).
+ * (Sebastian Kayser)
+ */
+# ifndef HAVE_ISWBLANK
+#  define iswblank(wc) iswctype(wc, wctype("blank"))
+# endif
+
 #else
 # define iswblank(c) (c == ' ')
 # define iswalnum(c) isalnum(c)
 # define iswprint(c) isprint(c)
-# undef UNICODE
-#endif
-
-#ifndef HAVE_NCURSESW_NCURSES_H
-# undef UNICODE
+# define towupper(c) toupper(c)
+# define towlower(c) tolower(c)
+# define iswalpha(c) isalpha(c)
 #endif
 
 extern int utf8_mode;
