@@ -1,9 +1,9 @@
 /*
  * utils.c      -- Various utility functions
  *
- * Copyright (C) 2005, 2006 Mikael Berthe <bmikael@lists.lilotux.net>
+ * Copyright (C) 2005-2008 Mikael Berthe <mikael@lilotux.net>
  * ut_* functions are derived from Cabber debug/log code.
- * from_iso8601() comes from the Gaim project.
+ * from_iso8601() comes from the Pidgin (libpurple) project.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -39,11 +39,27 @@
 #include <glib.h>
 
 #include <config.h>
+#include "utils.h"
 #include "logprint.h"
 
 static int DebugEnabled;
 static char *FName;
 
+//  expand_filename(filename)
+// Expand "~/" with the $HOME env. variable in a file name.
+// The caller must free the string after use.
+char *expand_filename(const char *fname)
+{
+  if (!fname)
+    return NULL;
+  if (!strncmp(fname, "~/", 2)) {
+    char *homedir = getenv("HOME");
+    if (homedir)
+      return g_strdup_printf("%s%s", homedir, fname+1);
+  }
+  return g_strdup(fname);
+}
+
 void ut_InitDebug(int level, const char *filename)
 {
   FILE *fp;
@@ -57,16 +73,13 @@ void ut_InitDebug(int level, const char *filename)
   }
 
   if (filename)
-    FName = g_strdup(filename);
+    FName = expand_filename(filename);
   else {
     FName = getenv("HOME");
     if (!FName)
-      FName = "/tmp/mcabberlog";
+      FName = g_strdup("/tmp/mcabberlog");
     else {
-      char *tmpname = g_new(char, strlen(FName) + 12);
-      strcpy(tmpname, FName);
-      strcat(tmpname, "/mcabberlog");
-      FName = tmpname;
+      FName = g_strdup_printf("%s/mcabberlog", FName);
     }
   }
 
@@ -92,8 +105,7 @@ void ut_InitDebug(int level, const char *filename)
   }
   fchmod(fileno(fp), S_IRUSR|S_IWUSR);
 
-  fprintf(fp, "New trace log started.\n"
-	  "----------------------\n");
+  fprintf(fp, "New trace log started.\n----------------------\n");
   fclose(fp);
 }
 
@@ -108,7 +120,8 @@ void ut_WriteLog(unsigned int flag, const char *data)
       scr_LogPrint(LPRINT_NORMAL, "ERROR: Cannot open tracelog file");
       return;
     }
-    fputs(data, fp);
+    if (fputs(data, fp) == EOF)
+      scr_LogPrint(LPRINT_NORMAL, "ERROR: Cannot write to tracelog file");
     fclose(fp);
   }
 }
@@ -123,11 +136,11 @@ int checkset_perm(const char *name, unsigned int setmode)
   struct stat buf;
 
 #ifdef __CYGWIN__
-  // Permission checking isn't efficent on Cygwin
+  // Permission checking isn't efficient on Cygwin
   return 0;
 #endif
 
-  fd = lstat(name, &buf);
+  fd = stat(name, &buf);
   if (fd == -1) return -1;
 
   if (buf.st_uid != geteuid()) {
@@ -191,14 +204,14 @@ int to_iso8601(char *dststr, time_t timestamp)
   tm_time = gmtime(&timestamp);
 
   ret = snprintf(dststr, 19, "%.4d%02d%02dT%02d:%02d:%02dZ",
-        1900+tm_time->tm_year, tm_time->tm_mon+1, tm_time->tm_mday,
+        (int)(1900+tm_time->tm_year), tm_time->tm_mon+1, tm_time->tm_mday,
         tm_time->tm_hour, tm_time->tm_min, tm_time->tm_sec);
 
   return ((ret == -1) ? -1 : 0);
 }
 
 //  from_iso8601(timestamp, utc)
-// This function came from the Gaim project, gaim_str_to_time().
+// This function came from the Pidgin project, gaim_str_to_time().
 // (Actually date may not be pure iso-8601)
 // Thanks, guys!
 // ** Modified by somian 10 Apr 2006 with advice from ysth.
@@ -210,6 +223,7 @@ time_t from_iso8601(const char *timestamp, int utc)
   char *c;
   int tzoff = 0;
   int hms_succ = 0;
+  int tmpyear;
 
   time(&retval);
   localtime_r(&retval, &t);
@@ -221,7 +235,8 @@ time_t from_iso8601(const char *timestamp, int utc)
   c = buf;
 
   /* 4 digit year */
-  if (!sscanf(c, "%04d", &t.tm_year)) return 0;
+  if (!sscanf(c, "%04d", &tmpyear)) return 0;
+  t.tm_year = tmpyear;
   c+=4;
   if (*c == '-')
     c++;
@@ -289,15 +304,6 @@ time_t from_iso8601(const char *timestamp, int utc)
   return retval;
 }
 
-// Should only be used for delays < 1s
-inline void safe_usleep(unsigned int usec)
-{
-  struct timespec req;
-  req.tv_sec = 0;
-  req.tv_nsec = (long)usec * 1000L;
-  nanosleep(&req, NULL);
-}
-
 /**
  * Derived from libjabber/jid.c, because the libjabber version is not
  * really convenient for our usage.
@@ -305,28 +311,28 @@ inline void safe_usleep(unsigned int usec)
  * Check if the full JID is valid
  * Return 0 if it is valid, non zero otherwise
  */
-int check_jid_syntax(char *jid)
+int check_jid_syntax(const char *fjid)
 {
-  char *str;
-  char *domain, *resource;
+  const char *str;
+  const char *domain, *resource;
   int domlen;
 
-  if (!jid) return 1;
+  if (!fjid) return 1;
 
-  domain = strchr(jid, '@');
+  domain = strchr(fjid, JID_DOMAIN_SEPARATOR);
 
   /* the username is optional */
   if (!domain) {
-    domain = jid;
+    domain = fjid;
   } else {
     /* node identifiers may not be longer than 1023 bytes */
-    if ((domain == jid) || (domain-jid > 1023))
+    if ((domain == fjid) || (domain-fjid > 1023))
       return 1;
     domain++;
 
     /* check for low and invalid ascii characters in the username */
-    for (str = jid; *str != '@'; str++) {
-      if (*str <= 32 || *str == ':' || *str == '@' ||
+    for (str = fjid; *str != JID_DOMAIN_SEPARATOR; str++) {
+      if (*str <= ' ' || *str == ':' || *str == JID_DOMAIN_SEPARATOR ||
               *str == '<' || *str == '>' || *str == '\'' ||
               *str == '"' || *str == '&') {
         return 1;
@@ -335,7 +341,7 @@ int check_jid_syntax(char *jid)
     /* the username is okay as far as we can tell without LIBIDN */
   }
 
-  resource = strchr(domain, '/');
+  resource = strchr(domain, JID_RESOURCE_SEPARATOR);
 
   /* the resource is optional */
   if (resource) {
@@ -355,7 +361,7 @@ int check_jid_syntax(char *jid)
   if (domlen > 1023) return 1;
 
   /* make sure the hostname is valid characters */
-  for (str = domain; *str != '\0' && *str != '/'; str++) {
+  for (str = domain; *str != '\0' && *str != JID_RESOURCE_SEPARATOR; str++) {
     if (!(isalnum(*str) || *str == '.' || *str == '-' || *str == '_'))
       return 1;
   }
@@ -365,7 +371,7 @@ int check_jid_syntax(char *jid)
 }
 
 
-void mc_strtolower(char *str)
+inline void mc_strtolower(char *str)
 {
   if (!str) return;
   for ( ; *str; str++)
@@ -382,26 +388,24 @@ void strip_arg_special_chars(char *s)
   int escape = FALSE;
   char *p;
 
+  if (!s) return;
+
   for (p = s; *p; p++) {
     if (*p == '"') {
       if (!escape) {
         instring = !instring;
         strcpy(p, p+1);
         p--;
-      } else {
+      } else
         escape = FALSE;
-      }
     } else if (*p == '\\') {
       if (!escape) {
-        if (*(p+1) == '"') {
-          strcpy(p, p+1);
-          p--;
-        }
-        escape = TRUE;
-      } else {
-        escape = FALSE;
+        strcpy(p, p+1);
+        p--;
       }
-    }
+      escape = !escape;
+    } else
+      escape = FALSE;
   }
 }
 
@@ -409,7 +413,7 @@ void strip_arg_special_chars(char *s)
 // Split the string arg into a maximum of n pieces, taking care of
 // double quotes.
 // Return a null-terminated array of strings.  This array should be freed
-// be the caller after use, for example with free_arg_lst().
+// by the caller after use, for example with free_arg_lst().
 // If dontstriplast is true, the Nth argument isn't stripped (i.e. no
 // processing of quote chars)
 char **split_arg(const char *arg, unsigned int n, int dontstriplast)
@@ -444,7 +448,7 @@ char **split_arg(const char *arg, unsigned int n, int dontstriplast)
       *(arglst+i) = g_strndup(start, p-start);
       strip_arg_special_chars(*(arglst+i));
       for (start = p+1; *start && *start == ' '; start++) ;
-      p = start;
+      p = start-1;
       i++;
     }
   }
@@ -483,32 +487,39 @@ void replace_nl_with_dots(char *bufstr)
 }
 
 //  ut_expand_tabs(text)
-// Expand tabs in string text.
-// If there is no tab in the string, a pointer to text is returned (be
-// careful _not_ to free the pointer in this case).
-// If there are some tabs, a new string with expanded chars is returned; this
-// is up to the caller to free this string after use.
+// Expand tabs and filter out some bad chars in string text.
+// If there is no tab and no bad chars in the string, a pointer to text
+// is returned (be careful _not_ to free the pointer in this case).
+// If there are some tabs or bad chars, a new string with expanded chars
+// and no bad chars is returned; this is up to the caller to free this
+// string after use.
 char *ut_expand_tabs(const char *text)
 {
-  char *xtext;
+  char *xtext, *linestart;
   char *p, *q;
-  guint8 n=0;
+  guint n = 0, bc = 0;
 
   xtext = (char*)text;
   for (p=xtext; *p; p++)
-    if (*p == '\t') n++;
+    if (*p == '\t')
+      n++;
+    else if (*p == '\x0d')
+      bc++;
+  // XXX Are there other special chars we should filter out?
 
-  if (!n)
+  if (!n && !bc)
     return (char*)text;
 
   xtext = g_new(char, strlen(text) + 1 + 8*n);
   p = (char*)text;
-  q = xtext;
+  q = linestart = xtext;
   do {
     if (*p == '\t') {
-      do { *q++ = ' '; } while ((q-xtext)%8);
-    } else {
+      do { *q++ = ' '; } while ((q-linestart)%8);
+    } else if (*p != '\x0d') {
       *q++ = *p;
+      if (*p =='\n')
+        linestart = q;
     }
   } while (*p++);
 
@@ -630,4 +641,15 @@ strcasestr (const char *haystack, const char *needle)
 }
 #endif /* !HAVE_STRCASESTR */
 
+//  startswith(str, word, ignore_case)
+// Returns TRUE if string str starts with word.
+int startswith(const char *str, const char *word, guint ignore_case)
+{
+  if (ignore_case && !strncasecmp(str, word, strlen(word)))
+    return TRUE;
+  else if (!ignore_case && !strncmp(str, word, strlen(word)))
+    return TRUE;
+  return FALSE;
+}
+
 /* vim: set expandtab cindent cinoptions=>2\:2(0:  For Vim users... */
